if [ -z ${CONDA_BUILD+x} ]; then
    source /opt/concourse/worker/volumes/live/e23e2ee9-1133-45b1-49fd-8099087467c2/volume/pymongo_1606344856863/work/build_env_setup.sh
fi
python setup.py install --single-version-externally-managed --record=record.txt